#  Copyright 2020 United Kingdom Research and Innovation
#  Copyright 2020 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from cil.framework import DataProcessor
from cil.processors.CofR_xcorrelation import CofR_xcorrelation
from cil.processors.CofR_image_sharpness import CofR_image_sharpness


class CentreOfRotationCorrector(DataProcessor):
    """
    This class contains methods to create a CentreOfRotationCorrector processor using the desired algorithm.
    """

    @staticmethod
    def xcorrelation(slice_index='centre', projection_index=0, ang_tol=0.1):
        r'''This creates a CentreOfRotationCorrector processor using the cross-correlation algorithm.

        For use on parallel-beam geometry it requires two projections 180 degree apart.

        Parameters
        ----------
        slice_index: int or str, optional
            An integer defining the vertical slice to run the algorithm on or string='centre' specifying the central slice should be used (default is 'centre')

        projection_index: int or list/tuple of ints, optional
            An integer defining the index of the first projection the cross correlation algorithm will use, where the second projection is chosen as the projection closest to 180 degrees from this.
            Or a list/tuple of ints specifying the two indices to be used for cross correlation (default is 0)

        ang_tol: float, optional
            The angular tolerance in degrees between the two input projections 180 degree gap (default is 0.1)

        Example
        -------
        >>> from cil.processors import CentreOfRotationCorrector
        >>> processor = CentreOfRotationCorrector.xcorrelation('centre')
        >>> processor.set_input(data)
        >>> data_centred = processor.get_ouput()

        Example
        -------
        >>> from cil.processors import CentreOfRotationCorrector
        >>> processor = CentreOfRotationCorrector.xcorrelation(slice_index=120)
        >>> processor.set_input(data)
        >>> processor.get_ouput(out=data)


        Example
        -------
        >>> from cil.processors import CentreOfRotationCorrector
        >>> import logging
        >>> logging.basicConfig(level=logging.WARNING)
        >>> cil_log_level = logging.getLogger('cil.processors')
        >>> cil_log_level.setLevel(logging.DEBUG)

        >>> processor = CentreOfRotationCorrector.xcorrelation(slice_index=120)
        >>> processor.set_input(data)
        >>> data_centred = processor.get_output()


        Note
        ----
        setting logging to 'debug' will give you more information about the algorithm progress



        '''
        processor = CofR_xcorrelation(slice_index, projection_index, ang_tol)
        return processor


    @staticmethod
    def image_sharpness(slice_index='centre', backend='tigre', tolerance=0.005, search_range=None, initial_binning=None):
        """This creates a CentreOfRotationCorrector processor.

        The processor will find the centre offset by maximising the sharpness of a reconstructed slice.

        Can be used on single slice parallel-beam, and centre slice cone beam geometry. For use only with datasets that can be reconstructed with FBP/FDK.

        Parameters
        ----------

        slice_index : int, str, default='centre'
            An integer defining the vertical slice to run the algorithm on. The special case slice 'centre' is the default.

        backend : {'tigre', 'astra'}
            The backend to use for the reconstruction

        tolerance : float, default=0.005
            The tolerance of the fit in pixels, the default is 1/200 of a pixel. This is a stopping criteria, not a statement of accuracy of the algorithm.

        search_range : int
            The range in pixels to search either side of the panel centre. If `None` a quarter of the width of the panel is used.

        initial_binning : int
            The size of the bins for the initial search. If `None` will bin the image to a step corresponding to <128 pixels. The fine search will be on unbinned data.


        Example
        -------
        from cil.processors import CentreOfRotationCorrector

        processor = CentreOfRotationCorrector.image_sharpness('centre', 'tigre')
        processor.set_input(data)
        data_centred = processor.get_output()


        Example
        -------
        from cil.processors import CentreOfRotationCorrector

        processor = CentreOfRotationCorrector.image_sharpness(slice_index=120, 'astra')
        processor.set_input(data)
        processor.get_output(out=data)


        Note
        ----
        For best results data should be 360deg which leads to blurring with incorrect geometry.
        This method is unreliable on half-scan data with 'tuning-fork' style artifacts.

        """
        processor = CofR_image_sharpness(slice_index=slice_index, backend=backend, tolerance=tolerance, search_range=search_range, initial_binning=initial_binning)
        return processor



import unittest
import numpy as np
import warnings
import pywt

from cil.framework import ImageGeometry, VectorGeometry
from cil.optimisation.functions import L1Norm, L1Sparsity
from cil.optimisation.operators import WaveletOperator

from testclass import CCPiTestClass

def setUp(self):
    np.random.seed(1)
    print("Seed set")

class TestWavelets(CCPiTestClass):
    
    def test_wavelet_operator_init(self):
        m, n=11,13
        dg = ImageGeometry(voxel_num_x=m, voxel_num_y=n)
        W = WaveletOperator(dg)
        self.assertEqual(W.wname, 'haar')
        self.assertEqual(pywt.Wavelet(W.wname).filter_bank, W._wavelet.filter_bank)
        self.assertEqual(W.correlation, None)
        self.assertEqual(W.bnd_cond, 'symmetric')
        self.assertEqual(W._trueAdj, True )
        self.assertEqual(W.level,int(np.log2(min(n,m))))
        self.assertNumpyArrayEqual(np.array((n,m)), W.domain_geometry().shape)    
                
        #Check is_linear
        self.assertTrue(W.is_linear())
        
        #Check is_orthogonal
        self.assertEqual(W.is_orthogonal(), W._wavelet.orthogonal)
    
        dg = ImageGeometry(voxel_num_x=m, voxel_num_y=n)
        rg = ImageGeometry(voxel_num_x=n, voxel_num_y=m)
        with self.assertRaises(AttributeError):
            W = WaveletOperator(dg, rg)
        
        dg = ImageGeometry(voxel_num_x=m, voxel_num_y=n)
        
        with self.assertWarnsRegex(UserWarning, "Level value of 100 is too high"):
            WaveletOperator(dg, level=100)
        
                
        W = WaveletOperator(dg, dg, wname='db2', level=0, axes=[0], bnd_cond='zero', true_adjoint=False)
        self.assertEqual(W.wname, 'db2')
        self.assertEqual(pywt.Wavelet(W.wname).filter_bank, W._wavelet.filter_bank)
        self.assertEqual(W.correlation, None)
        self.assertEqual(W.bnd_cond, 'zero')
        self.assertEqual(W._trueAdj, False)
        self.assertEqual(W.level,0)
        self.assertNumpyArrayEqual(np.array((n,m)), W.domain_geometry().shape)
        
        
    def test_wavelet_norm_init(self):
        m, n=11,13
        dg = ImageGeometry(voxel_num_x=m, voxel_num_y=n)
        W = WaveletOperator(dg)
        Wnorm=L1Sparsity(W)
        from cil.optimisation.functions.L1Norm import _L1Norm, _WeightedL1Norm
        self.assertTrue(issubclass(Wnorm.l1norm.function.__class__, _L1Norm))
        
        Wnorm=L1Sparsity(W, weight=np.ones((13,11)))
        self.assertTrue(issubclass(Wnorm.l1norm.function.__class__, _WeightedL1Norm))
        self.assertNumpyArrayEqual(Wnorm.l1norm.function.weight, np.ones((13,11)))
    

    def test_WaveletOperator_dimensions(self):
        m, n = 42, 47
        level = 2
        for wname in ['haar', 'db2', 'db3', 'db4', 'sym2', 'sym3', 'coif2', 'coif3']:
            for dg in [ImageGeometry(voxel_num_x=m, voxel_num_y=n), VectorGeometry(m), VectorGeometry(n)]:
                with self.subTest(msg=f"{wname} transform failed for {dg.__class__.__name__} of size {dg.shape}", wname=wname, dg=dg):
                    x = dg.allocate('random')

                    W = WaveletOperator(dg, wname=wname, level=level)
                    rg = W.range_geometry() # Range
                    Wx = W.direct(x)
                    y = W.adjoint(Wx)

                    self.assertEqual(Wx.shape, rg.shape, msg="Coefficient array shape should match range_geometry")
                    self.assertEqual(y.shape, dg.shape, msg="Adjoint array shape should match domain_geometry")

                    # Reconstruction should be (almost) the original input
                    self.assertNumpyArrayAlmostEqual(x.as_array(), y.as_array())

                    self.assertRaises(AttributeError, WaveletOperator, domain_geometry=dg, range_geometry=dg, wname=wname, msg="Geometry shape mismatch should raise error")
    
    def test_WaveletOperator_complex_input(self):
        m, n = 42, 47
        level = 2
        for wname in ['haar', 'db2', 'db3', 'db4', 'sym2', 'sym3', 'coif2', 'coif3']:
            for dg in [ImageGeometry(voxel_num_x=m, voxel_num_y=n), VectorGeometry(m), VectorGeometry(n)]:
                with self.subTest(msg=f"{wname} transform failed for {dg.__class__.__name__} of size {dg.shape}", wname=wname, dg=dg):
                    x = dg.allocate('random', dtype='complex128')

                    W = WaveletOperator(dg, wname=wname, level=level)
                    rg = W.range_geometry() # Range
                    Wx = W.direct(x)
                    y = W.adjoint(Wx)

                    self.assertEqual(x.dtype, Wx.dtype, msg="Complex input should give complex coefficients")
                    self.assertEqual(Wx.dtype, y.dtype, msg="Complex coefficients should give complex output")

                    self.assertEqual(Wx.shape, rg.shape, msg="Coefficient array shape should match range_geometry")
                    self.assertEqual(y.shape, dg.shape, msg="Adjoint array shape should match domain_geometry")

                    # Reconstruction should be (almost) the original input
                    self.assertNumpyArrayAlmostEqual(x.as_array(), y.as_array())

                    z = 0*Wx
                    W.direct(x, out=z)
                    self.assertNumpyArrayAlmostEqual(Wx.as_array(), z.as_array())
    
    def test_WaveletOperator_dimensions_biorthogonal(self):
        m, n = 48, 47
        level = 2
        for wname in ["bior3.5", "bior3.3", "rbio3.5", "rbio4.4"]:
            for dg in [ImageGeometry(voxel_num_x=m, voxel_num_y=n), VectorGeometry(m), VectorGeometry(n)]:
                with self.subTest(msg=f"{wname} transform failed for {dg.__class__.__name__} of size {dg.shape}", wname=wname, dg=dg):
                    x = dg.allocate('random')

                    W = WaveletOperator(dg, wname=wname, level=level)
                    rg = W.range_geometry() # Range
                    Wx = W.direct(x)
                    y = W.adjoint(Wx)

                    self.assertEqual(Wx.shape, rg.shape, msg="Coefficient array shape should match range_geometry")
                    self.assertEqual(y.shape, dg.shape, msg="Adjoint array shape should match domain_geometry")
                    
                    #Check is_not_orthogonal
                    self.assertEqual(W.is_orthogonal(), False)

                    # Reconstruction should be imperfect
                    self.assertLess(0, (x - y).norm(), msg="Biorthogonal wavelets with true adjoint should no longer produce perfect reconstructions")

                    self.assertRaises(AttributeError, WaveletOperator, domain_geometry=dg, range_geometry=dg, wname=wname, msg="Geometry shape mismatch should raise error")
    
    def test_WaveletOperator_biorthogonal_reconstruction(self):
        m, n = 48, 49
        level = 2
        for wname in ["bior3.5", "bior3.3", "rbio3.5", "rbio4.4"]:
            for dg in [ImageGeometry(voxel_num_x=m, voxel_num_y=n), VectorGeometry(m), VectorGeometry(n)]:
                with self.subTest(msg=f"{wname} transform failed for {dg.__class__.__name__} of size {dg.shape}", wname=wname, dg=dg):
                    x = dg.allocate('random')

                    W = WaveletOperator(dg, wname=wname, level=level, true_adjoint=False)
                    rg = W.range_geometry() # Range
                    Wx = W.direct(x)
                    y = W.adjoint(Wx)

                    self.assertEqual(Wx.shape, rg.shape, msg="Coefficient array shape should match range_geometry")
                    self.assertEqual(y.shape, dg.shape, msg="Adjoint array shape should match domain_geometry")

                    # Reconstruction should be (almost) the original input
                    self.assertNumpyArrayAlmostEqual(x.as_array(), y.as_array())

                    self.assertRaises(AttributeError, WaveletOperator, domain_geometry=dg, range_geometry=dg, wname=wname, msg="Geometry shape mismatch should raise error")

    def test_wavelet_axes(self):
        m, n = 20, 21
        dg = ImageGeometry(voxel_num_x=m, voxel_num_y=n) # Domain
        dg.channels = 32

        # Give axes as range
        W = WaveletOperator(dg, wname='db2', level=2, axes=range(1,3))
        rg = W.range_geometry()

        self.assertEqual(dg.shape[0], rg.shape[0], msg="Size of channels should remain unchanged")
        self.assertLess(dg.shape[1:], rg.shape[1:], msg="Decomposed dimensions should get extended")

        # Try each individual axis
        for ax in range(3):
            with self.subTest(msg=f"Failed for axis {ax}", ax=ax):
                W = WaveletOperator(dg, wname='db2', level=2, axes=[ax])
                rg = W.range_geometry()

                self.assertEqual(dg.shape[not ax], rg.shape[not ax], msg="Size of other dimensions should remain unchanged")
                self.assertEqual(dg.shape[ax] + 5, rg.shape[ax], msg="Decomposed dimension should get extended by 5 (when level=2)")

    def test_wavelet_correlation(self):
        m, n = 20, 21
        dg = ImageGeometry(voxel_num_x=m, voxel_num_y=n) # Domain
        
        self.assertRaises(AttributeError, WaveletOperator, dg, wname='haar', correlation='channel', msg="Missing channels should raise error")

        dg.channels = 32
        self.assertRaises(AttributeError, WaveletOperator, dg, wname='haar', correlation='SPACEchannel', msg="Invalid correlation keyword should raise error")

        W = WaveletOperator(dg, wname='db2', level=2, correlation='Space')
        rg = W.range_geometry()
        self.assertEqual(dg.shape[0], rg.shape[0], msg="Size of channels should remain unchanged")
        self.assertLess(dg.shape[1:], rg.shape[1:], msg="Correlated dimensions should get extended")

        self.assertWarns(UserWarning, WaveletOperator, dg, axes=[0], correlation='channel', msg="Defining both axes and correlation should give warning")

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            W = WaveletOperator(dg, wname='db2', level=2, axes=[1], correlation='channel')
        rg = W.range_geometry()

        self.assertEqual(dg.shape[0], rg.shape[0], msg="Axes should be prioritized over correlation")

    def test_wavelet_bnd_cond(self):
        m, n = 20, 24
        dg = ImageGeometry(voxel_num_x=m, voxel_num_y=n) # Domain
        W = WaveletOperator(dg, wname='db2', level=2, bnd_cond='periodization') # Note: this is different from bnd_cond='periodic'
        rg = W.range_geometry()

        self.assertEqual(dg.shape, rg.shape, msg="Periodization convolution should require no padding")

        x = dg.allocate('random')
        c = rg.allocate('random')

        ip1 = c.dot(W.direct(x))
        ip2 = x.dot(W.adjoint(c))
        M = x.norm() # Normalization
        self.assertAlmostEqual(ip1/M, ip2/M, places=5, msg="Periodic convolution should be closest to true adjoint")

    def test_wavelet_adjoint(self):
        m, n = 48, 64
        dg = ImageGeometry(voxel_num_x=m, voxel_num_y=n) # Domain
        x = dg.allocate('random')
        for wname in ['haar', 'db2', 'db3', 'db4', 'sym2', 'sym3', 'coif2', 'coif3', 'bior3.5', 'bior3.3', 'rbio3.5', 'rbio3.3']:
            with self.subTest(msg=f"Failed for wavelet {wname}", wname=wname):
                W = WaveletOperator(dg, wname=wname, level=2, bnd_cond='periodization') # Note: this is different from bnd_cond='periodic'
                rg = W.range_geometry()
                c = rg.allocate('random')

                ip1 = c.dot(W.direct(x))
                ip2 = x.dot(W.adjoint(c))
                M = x.norm() # Normalization
                self.assertAlmostEqual(ip1/M, ip2/M, places=5, msg="Periodization convolution should be closest to true adjoint")
    
    def test_WaveletOperator_norm(self):
        n = 64
        dg = VectorGeometry(n)
        for wname in ['haar', 'db2', 'db3', 'db4', 'sym2', 'sym3', 'coif2', 'coif3']:
            with self.subTest(msg=f"Failed for wavelet {wname}", wname=wname):
                W = WaveletOperator(dg, wname=wname, level=1)
                self.assertEqual(W.norm(), 1.0, msg="Orthogonal wavelet transform should have unit norm")
        
        for wname in ['bior3.5', 'bior3.3', 'rbio3.5', 'rbio3.3']:
            with self.subTest(msg=f"Failed for wavelet {wname}", wname=wname):
                W = WaveletOperator(dg, wname=wname, level=2, bnd_cond='periodization') # Need to set bnd_cond for most faithful adjoint
                self.assertLess(0.9, W.norm())

    def test_L1Sparsity(self):
        n = 20
        wname = 'db2'
        level = 2

        dg = VectorGeometry(n)
        W = WaveletOperator(dg, wname=wname, level=level)
        WN = L1Sparsity(W)

        x = dg.allocate('random')
        Wx = W.direct(x)

        self.assertAlmostEqual(WN(x), np.sum(Wx.abs().as_array()), msg="L1Sparsity should be the sum of absolute values of the wavelet coefficients", places=5)

        y = W.adjoint(Wx)
        tau = 0.0
        prox = WN.proximal(x, tau=tau)

        self.assertAlmostEqual(0, (prox - y).norm(), places=5, msg="Prox_{tau=0}(  ) should be (almost) the identity")
        self.assertNumpyArrayAlmostEqual(y.as_array(), prox.as_array())

        tau = 1.1
        Wx /= Wx.abs().max() # Bound Wx to [-1, 1]
        W.adjoint(Wx, out=y)
        WN.proximal(y, tau=tau, out=prox)
        self.assertNumpyArrayAlmostEqual(np.zeros(x.shape), prox.as_array())

        # Test for sign problems in prox
        prox = x.copy()
        WN.proximal(prox, tau=tau, out=prox)
        proxNeg = WN.proximal(-x, tau=tau)
        self.assertNumpyArrayAlmostEqual(np.zeros(x.shape), (prox + proxNeg).as_array())

        convConj = WN.convex_conjugate(1.2*y)
        self.assertEqual(np.inf, convConj, msg="Some coefficient should be > 1")

        convConj = WN.convex_conjugate(0.9*y)
        self.assertEqual(0, convConj, msg="All coefficient should be < 1")

        dg = VectorGeometry(48)
        for true_adjoint in [True, False]:
                W = WaveletOperator(dg, wname='bior3.5', level=1, true_adjoint=true_adjoint)
                self.assertEqual(W.is_orthogonal(), False)
                self.assertWarns(UserWarning, L1Sparsity, W)


    
    def test_L1Sparsity_complex_input(self):
        n = 20
        wname = 'db2'
        level = 2

        dg = VectorGeometry(n)
        W = WaveletOperator(dg, wname=wname, level=level)
        self.assertEqual(W.is_orthogonal(), True)
        WN = L1Sparsity(W)

        x = dg.allocate('random', dtype='complex64')
        Wx = W.direct(x)

        self.assertAlmostEqual(WN(x), np.sum(Wx.abs().as_array()), msg="L1Sparsity should be the sum of absolute values of the wavelet coefficients", places =5 )

        y = W.adjoint(Wx)
        tau = 0.0
        prox = WN.proximal(x, tau=tau)

        self.assertAlmostEqual(0, (prox - y).norm(), places=5, msg="Prox_{tau=0}(  ) should be (almost) the identity")
        self.assertNumpyArrayAlmostEqual(y.as_array(), prox.as_array())

        tau = 1.1
        Wx /= Wx.abs().max() # Bound Wx to [-1, 1]
        W.adjoint(Wx, out=y)
        WN.proximal(y, tau=tau, out=prox)
        self.assertNumpyArrayAlmostEqual(np.zeros(x.shape), prox.as_array())

        # Test for sign problems in prox
        prox = x.copy()
        WN.proximal(prox, tau=tau, out=prox)
        proxNeg = WN.proximal(-x, tau=tau)
        self.assertNumpyArrayAlmostEqual(np.zeros(x.shape), (prox + proxNeg).as_array())

        convConj = WN.convex_conjugate(1.2*y)
        self.assertEqual(np.inf, convConj, msg="Some coefficient should be > 1")

        convConj = WN.convex_conjugate(0.9*y)
        self.assertEqual(0, convConj, msg="All coefficient should be < 1")


if __name__ == "__main__":
    unittest.main()


#   This work is part of the Core Imaging Library (CIL) developed by CCPi
#   (Collaborative Computational Project in Tomographic Imaging), with
#   substantial contributions by UKRI-STFC and University of Manchester.

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import numpy as np
import math
from functools import partial
import time
import numbers

class Sampler():
    # TODO: Work out how to make the examples testable
    """
    Initialises a sampler that returns and then increments indices from a sequence defined by a function.

    Static methods to easily configure several samplers are provided, such as sequential, staggered, Herman-Mayer, random with and without replacement.




    Custom deterministic samplers can be created by using the `from_function` static method or by subclassing this sampler class.



    Parameters
    ----------

    function : Callable[[int], int]
        A function that takes an integer iteration number and returns an integer between 0 and num_indices.

    num_indices: int
        The sampler will select from a range of indices 0 to num_indices.

    sampling_type:str, optional,  default = None
        The sampling type used. This is recorded for reference and printed when `print` is called.

    prob_weights: list of floats of length num_indices that sum to 1.  Default is [1 / num_indices] * num_indices
        Consider that the sampler is incremented a large number of times this argument holds the expected number of times each index would be outputted,  normalised to 1.

    Returns
    -------
    Sampler
        An instance of the Sampler class representing the desired configuration.

    Example
    -------
    >>> sampler = Sampler.random_with_replacement(5)
    >>> print(sampler.get_samples(20))
    [3 4 0 0 2 3 3 2 2 1 1 4 4 3 0 2 4 4 2 4]
    >>> print(next(sampler))
    3
    >>> print(sampler.next())
    4


    >>> sampler = Sampler.staggered(num_indices=21, stride=4)
    >>> print(next(sampler))
    0
    >>> print(sampler.next())
    4
    >>> print(sampler.get_samples(5))
    [ 0  4  8 12 16]

    Example
    -------
    >>> sampler = Sampler.sequential(10)
    >>> print(sampler.get_samples(5))
    >>> print(next(sampler))
    0
    [0 1 2 3 4]
    >>> print(sampler.next())
    1

    Example
    -------
    >>> sampler = Sampler.herman_meyer(12)
    >>> print(sampler.get_samples(16))
    [ 0  6  3  9  1  7  4 10  2  8  5 11  0  6  3  9]



    Example
    --------
    This example creates a sampler that outputs sequential indices, starting from 1.

    >>> num_indices=10
    >>>
    >>> def my_sampling_function(iteration_number):
    >>>     return (iteration_number+1)%10
    >>>
    >>> sampler = Sampler.from_function(num_indices=num_indices, function=my_sampling_function)
    >>> print(list(sampler.get_samples(25)))
    [1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 1, 2, 3, 4, 5]



    Note
    -----
    The optimal choice of sampler depends on the data and the number of calls to the sampler.  Note that a low number of calls to a random sampler won't give an even distribution.
    For a small number of samples (e.g. `<5*num_indices`) the user may wish to consider another sampling method e.g. random without replacement, which, when calling `num_indices` samples is guaranteed to draw each index exactly once.
        """

    def __init__(self, num_indices, function,  sampling_type=None, prob_weights=None):

        self._type = sampling_type

        if isinstance (num_indices, numbers.Integral):
            self._num_indices = num_indices
        else:
            raise ValueError('`num_indices` should be an integer. ')

        if callable(function):
            self._function = function
        else:
            raise ValueError('`function` should be an callable function. ')

        if prob_weights is None:
            prob_weights = [1 / num_indices] * num_indices
        else:
            if abs(sum(prob_weights) - 1) > 1e-6:
                raise ValueError('The provided prob_weights must sum to one')

            if any(np.array(prob_weights) < 0):
                raise ValueError(
                    'The provided prob_weights must be greater than or equal to zero')

        self._prob_weights = prob_weights
        self._iteration_number = 0

    @property
    def prob_weights(self):
        return self._prob_weights

    @property
    def num_indices(self):
        return self._num_indices

    @property
    def current_iter_number(self):
        return self._iteration_number

    def next(self):
        """
        Returns a sample from the list of indices `{0, 1, …, N-1}, where N is the number of indices and increments the sampler.
        """

        out = self._function(self._iteration_number)

        self._iteration_number += 1
        return out

    def __next__(self):
        return self.next()

    def get_samples(self,  num_samples):
        """
        Generates a list of the first num_samples output by the sampler. Calling this does not increment the sampler index or affect the behaviour of the sampler .

        Parameters
        ----------
        num_samples: int
            The number of samples to return.

        Returns
        --------
        List
            The first `num_samples" output by the sampler.
        """
        save_last_index = self._iteration_number
        self._iteration_number = 0

        output = [self.next() for _ in range(num_samples)]

        self._iteration_number = save_last_index

        return np.array(output)

    def __str__(self):
        repres = "Sampler that selects from a list of indices {0, 1, …, N-1}, where N is the number of indices. \n"
        repres += "Type : {} \n".format(self._type)
        repres += "Current iteration number : {} \n".format(
            self._iteration_number)
        repres += "Number of indices : {} \n".format(self._num_indices)
        repres += "Probability weights : {} \n".format(self._prob_weights)
        return repres

    @staticmethod
    def sequential(num_indices):
        """
        Instantiates a sampler that outputs sequential indices.

        Parameters
        ----------
        num_indices: int
            The sampler will select from a range of indices 0 to num_indices.

        Returns
        -------
        Sampler
            An instance of the Sampler class that will generate indices sequentially.

        Example
        -------
        >>> sampler = Sampler.sequential(10)
        >>> print(sampler.get_samples(5))
        >>> print(next(sampler))
        0
        [0 1 2 3 4]
        >>> print(sampler.next())
        1
        """
        def function(x):
            return x % num_indices

        sampler = Sampler(function=function, num_indices=num_indices,
                          sampling_type='sequential'
                          )
        return sampler

    @staticmethod
    def _staggered_function(num_indices, stride, iter_number):
        """Function that takes in an iteration number and outputs an index number based on the staggered ordering.

        Parameters
        ----------
        num_indices: int
            The sampler will select from a range of indices 0 to num_indices.

        stride: int
            The stride between returned indices. The stride should be less than the num_indices.

        iter_number: int
            The current iteration number of the sampler.

        Returns
        -------
        int
            The index to be outputted by the sampler corresponding to the `iter_number`

        """
        if not isinstance (num_indices, numbers.Integral):
            raise ValueError('`num_indices` should be an integer. ')

        iter_number_mod = iter_number % num_indices
        floor = num_indices // stride
        mod = num_indices % stride

        if iter_number_mod < (floor + 1)*mod:
            row_number = iter_number_mod // (floor + 1)
            column_number = (iter_number_mod % (floor + 1))
        else:
            row_number = mod + (iter_number_mod - (floor+1)*mod) // floor
            column_number = (iter_number_mod - (floor+1)*mod) % floor

        return row_number + stride*column_number

    @staticmethod
    def staggered(num_indices, stride):
        """
        Instantiates a sampler which outputs in a staggered order.

        Parameters
        ----------
        num_indices: int
            The sampler will select from a range of indices 0 to num_indices.

        stride: int
            The stride between returned indices. The stride should be less than the num_indices.

        Returns
        -------
        Sampler
            An instance of the Sampler class that will generate indices in a staggered pattern.


        Example
        -------
        >>> sampler = Sampler.staggered(num_indices=21, stride=4)
        >>> print(next(sampler))
        0
        >>> print(sampler.next())
        4
        >>> print(sampler.get_samples(5))
        [ 0  4  8 12 16]
        Example
        -------
        >>> sampler = Sampler.staggered(num_indices=17, stride=8)
        >>> print(next(sampler))
        0
        >>> print(sampler.next())
        8
        >>> print(sampler.get_samples(10))
        [ 0  8  16 1 9 2 10 3 11 4]


        """

        if stride >= num_indices:
            raise (ValueError('The stride should be less than the number of indices'))

        sampler = Sampler(function=partial(Sampler._staggered_function, num_indices, stride),
                          num_indices=num_indices, sampling_type='staggered'
                          )

        return sampler

    @staticmethod
    def random_with_replacement(num_indices, prob=None, seed=None):
        """
        Instantiates a sampler which outputs an index between 0 - num_indices with a given probability.

        Parameters
        ----------
        num_indices: int
            The sampler will select from a range of indices 0 to num_indices

        prob: list of floats, optional
            The probability for each index to be selected by the 'next' operation. If not provided, the indices will be sampled uniformly. The list should have a length equal to num_indices, and the values should sum to 1

        seed:int, optional
            Used to initialise the random number generator where repeatability is required.

        Returns
        -------
        `RandomSampler`
            An instance of the `RandomSampler` class that will generate indices randomly with replacement

        Example
        -------
        >>> sampler = Sampler.random_with_replacement(5)
        >>> print(sampler.get_samples(10))
        [3 4 0 0 2 3 3 2 2 1]
        >>> print(next(sampler))
        3
        >>> print(sampler.next())
        4

        >>> sampler = Sampler.random_with_replacement(num_indices=4, prob=[0.7,0.1,0.1,0.1])
        >>> print(sampler.get_samples(10))
        [0 1 3 0 0 3 0 0 0 0]
        """

        sampler = SamplerRandom(
            num_indices=num_indices,
            sampling_type='random_with_replacement',
            prob=prob,
            replace=True,
            seed=seed
        )
        return sampler

    @staticmethod
    def random_without_replacement(num_indices, seed=None):
        """
        Instantiates a sampler which outputs an index between 0 - num_indices. Once sampled the index will not be sampled again until all indices have been returned.

        Parameters
        ----------
        num_indices: int
            The sampler will select from a range of indices 0 to num_indices.

        seed: int, optional
            Used to initialise the random number generator where repeatability is required.

        Returns
        -------
        `RandomSampler`
            An instance of the `RandomSampler` class that will generate indices randomly without replacement

        Example
        -------
        >>> sampler=Sampler.randomWithoutReplacement(num_indices=7, seed=1)
        >>> print(sampler.get_samples(16))
        [6 2 1 0 4 3 5 1 0 4 2 5 6 3 3 2]

        """

        sampler = SamplerRandom(
            num_indices=num_indices,
            sampling_type='random_without_replacement',
            replace=False,
            seed=seed
        )
        return sampler

    @staticmethod
    def from_function(num_indices, function, prob_weights=None):
        """
        Instantiate a sampler that wraps a function for index selection.

        Parameters
        ----------
        num_indices: int
            The sampler will select from a range of indices 0 to num_indices.

    function : callable
        A deterministic function that takes an integer as an argument, representing the iteration number, and returns an integer between 0 and num_indices. The function signature should be function(iteration_number: int) -> int

        prob_weights: list of floats of length num_indices that sum to 1. Default is [1 / num_indices] * num_indices
            Consider that the sampler is incremented a large number of times this argument holds the expected number of times each index would be outputted,  normalised to 1.

        Returns
        -------
        Sampler
            An instance of the Sampler class which samples from a function.


        Example
        --------
        This example creates a sampler that always outputs 2.  The probability weights are passed to the sampler as they are not uniform.

        >>> num_indices=3
        >>>
        >>> def my_sampling_function(iteration_number):
        >>>     return 2
        >>>
        >>> sampler = Sampler.from_function(num_indices=num_indices, function=my_sampling_function, prob_weights=[0, 0, 1])
        >>> print(list(sampler.get_samples(12)))
        [2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2]


        Example
        --------
        This example creates a sampler that outputs sequential indices, starting from 1.  The probability weights are not passed to the sampler as they are uniform.

        >>> num_indices=10
        >>>
        >>> def my_sampling_function(iteration_number):
        >>>     return (iteration_number+1)%10
        >>>
        >>> sampler = Sampler.from_function(num_indices=num_indices, function=my_sampling_function)
        >>> print(list(sampler.get_samples(25)))
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 1, 2, 3, 4, 5]


        Example
        -------
        This example creates a sampler that samples in order from a custom list. The num_indices  is 6, although note that the index 5 is never output by the sampler. The number of indices must be at least one greater than any of the elements in the custom_list.
        The probability weights are passed to the sampler as they are not uniform.

        >>> custom_list = [0,0,0,0,0,0,3,2,1,4]
        >>> num_indices = 6
        >>>
        >>> def my_sampling_function(iteration_number, custom_list=custom_list]):
        >>>    return(custom_list[iteration_number%len(custom_list)])
        >>>
        >>> sampler = Sampler.from_function(num_indices=num_indices, function=my_sampling_function, prob_weights=[0.6, 0.1, 0.1, 0.1, 0.1, 0.0])
        >>> print(list(sampler.get_samples(25)))
        [0, 0, 0, 0, 0, 0, 3, 2, 1, 4, 0, 0, 0, 0, 0, 0, 3, 2, 1, 4, 0, 0, 0, 0, 0]
        >>> print(sampler)
        Sampler that wraps a function that takes an iteration number and selects from a list of indices {0, 1, …, N-1}, where N is the number of indices.
        Type : from_function
        Current iteration number : 0
        number of indices : 6
        Probability weights : [0.6, 0.1, 0.1, 0.1, 0.1, 0.0]

        """

        sampler = Sampler(
            num_indices=num_indices,
            sampling_type='from_function',
            function=function,
            prob_weights=prob_weights
        )
        return sampler

    @staticmethod
    def _prime_factorisation(n):
        """
        Parameters
        ----------

        n: int
            The number to be factorised.

        Returns
        -------
        list of ints
            The prime factors of n.

        """
        factors = []

        while n % 2 == 0:
            n //= 2
            factors.append(2)

        i = 3
        while i*i <= n:
            while n % i == 0:
                n //= i
                factors.append(i)
            i += 2

        if n > 1:
            factors.append(n)

        return factors

    @staticmethod
    def _herman_meyer_function(num_indices,  addition_arr, repeat_length_arr, iteration_number):
        """
        Parameters
        ----------
        num_indices: int
            The number of indices to be sampled from.

        addition_arr: list of ints
            The product of all factors at indices greater than the current factor.

        repeat_length_arr: list of ints
            The product of all factors at indices less than the current factor.

        iteration_number: int
            The current iteration number.

        Returns
        -------
        int
            The index to be sampled from.

        """

        index = 0
        for n in range(len(addition_arr)):
            addition = addition_arr[n]
            repeat_length = repeat_length_arr[n]

            length = num_indices // (addition*repeat_length)
            arr = np.arange(length) * addition

            ind = math.floor(iteration_number/repeat_length) % length
            index += arr[ind]

        return index

    @staticmethod
    def herman_meyer(num_indices):
        """
        Instantiates a sampler which outputs in a Herman Meyer order.

        Parameters
        ----------
        num_indices: int
            The sampler will select from a range of indices 0 to num_indices. For Herman-Meyer sampling this number should not be prime.

        Reference
        ----------
        With thanks to Imraj Singh and Zeljko Kereta for their help with the initial implementation of the Herman Meyer sampling. Their implementation was used in:

        Singh I, et al. Deep Image Prior PET Reconstruction using a SIRF-Based Objective - IEEE MIC, NSS & RTSD 2022. https://discovery.ucl.ac.uk/id/eprint/10176077/1/MIC_Conference_Record.pdf

        The sampling method was introduced in:

        Herman GT, Meyer LB. Algebraic reconstruction techniques can be made computationally efficient. IEEE Trans Med Imaging.  doi: 10.1109/42.241889.

        Returns
        -------
        Sampler
            An instance of the Sampler class which outputs in a Herman Meyer order.

        Example
        -------
        >>> sampler=Sampler.herman_meyer(12)
        >>> print(sampler.get_samples(16))
        [ 0  6  3  9  1  7  4 10  2  8  5 11  0  6  3  9]
        """

        if not isinstance (num_indices, numbers.Integral):
            raise ValueError('`num_indices` should be an integer. ')

        factors = Sampler._prime_factorisation(num_indices)

        n_factors = len(factors)
        if n_factors == 1:
            raise ValueError(
                'Herman Meyer sampling defaults to sequential ordering if the number of indices is prime. Please use an alternative sampling method or change the number of indices. ')

        addition_arr = np.empty(n_factors, dtype=np.int64)
        repeat_length_arr = np.empty(n_factors, dtype=np.int64)

        repeat_length = 1
        addition = num_indices
        for i in range(n_factors):
            addition //= factors[i]
            addition_arr[i] = addition

            repeat_length_arr[i] = repeat_length
            repeat_length *= factors[i]

        hmf_call = partial(Sampler._herman_meyer_function,
                           num_indices, addition_arr, repeat_length_arr)

        # define the sampler
        sampler = Sampler(function=hmf_call,
                          num_indices=num_indices,
                          sampling_type='herman_meyer',
                          prob_weights=[1 / num_indices] * num_indices
                          )

        return sampler


class SamplerRandom(Sampler):
    """
    The user is recommended to not instantiate this class  directly but instead use one of the static methods  in the parent Sampler class that will return instances of different samplers.

    This class produces Samplers that output random samples with and without replacement from the set {0, 1, …, N-1} where N=num_indices.

    Custom random samplers can be created by subclassing this sampler class.

    Parameters
    ----------

    num_indices: int
        The sampler will select from a range of indices 0 to num_indices.

    sampling_type:str, optional,  default = 'random_with_replacement"
        The sampling type used. This is recorded for reference and printed when `print` is called.

    prob_weights: list of floats of length num_indices that sum to 1.  Default is [1 / num_indices] * num_indices
        Consider that the sampler is incremented a large number of times this argument holds the expected number of times each index would be outputted,  normalised to 1.

    replace: bool, default is True
        If True, sample with replace, otherwise sample without replacement

    seed:int, optional
        Used to initialise the random number generator where repeatability is required.

    Returns
    -------
    Sampler
        An instance of the Sampler class representing the desired configuration.

    Example
    -------
    >>> sampler = Sampler.random_with_replacement(5)
    >>> print(sampler.get_samples(20))
    [3 4 0 0 2 3 3 2 2 1 1 4 4 3 0 2 4 4 2 4]

    Example
    -------
    >>> sampler=Sampler.randomWithoutReplacement(num_indices=7, seed=1)
    >>> print(sampler.get_samples(16))
    [6 2 1 0 4 3 5 1 0 4 2 5 6 3 3 2]

    """

    def __init__(self, num_indices,  seed=None, replace=True, prob=None,  sampling_type='random_with_replacement'):

        if seed is not None:
            self._seed = seed
        else:
            self._seed = int(time.time())
        self._generator = np.random.RandomState(self._seed)
        self._sampling_list = None
        self._replace = replace

        super(SamplerRandom, self).__init__(num_indices, self._function,
                                            sampling_type=sampling_type, prob_weights=prob)

    @property
    def seed(self):
        return self._seed

    @property
    def replace(self):
        return self._replace

    def _function(self, iteration_number):
        """ For each iteration number this function samples from a randomly generated list in order. Every num_indices the list is re-created. """
        location = iteration_number % self._num_indices
        if location == 0:
            self._sampling_list = self._generator.choice(
                self._num_indices, self._num_indices, p=self._prob_weights, replace=self._replace)
        out = self._sampling_list[location]
        return out

    def get_samples(self,  num_samples):
        """
        Generates a list of the first num_samples output by the sampler. Calling this does not increment the sampler index or affect the behaviour of the sampler .

        Parameters
        ----------
        num_samples: int
            The number of samples to return.
        Returns
        -------
        list
            The first `num_samples` produced by the sampler
        """
        save_last_index = self._iteration_number
        self._iteration_number = 0

        save_generator = self._generator
        self._generator = np.random.RandomState(self._seed)
        save_sampling_list = self._sampling_list

        output = [self.next() for _ in range(num_samples)]

        self._iteration_number = save_last_index

        self._generator = save_generator
        self._sampling_list = save_sampling_list

        return np.array(output)

    def __str__(self):
        repres = super().__str__()
        repres += "Seed : {} \n".format(self._seed)
        return repres



#  Copyright 2019 United Kingdom Research and Innovation
#  Copyright 2019 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

import unittest
from unittest.mock import Mock
from utils import initialise_tests
from cil.framework import ImageGeometry, BlockGeometry, VectorGeometry, BlockDataContainer, DataContainer
from cil.optimisation.operators import BlockOperator,\
    FiniteDifferenceOperator, SymmetrisedGradientOperator
import numpy
from timeit import default_timer as timer
from cil.optimisation.operators import GradientOperator, IdentityOperator,\
    DiagonalOperator, MaskOperator, ChannelwiseOperator, BlurringOperator
from cil.optimisation.operators import LinearOperator, MatrixOperator
import numpy
from cil.optimisation.operators import SumOperator,  ZeroOperator, CompositionOperator, ProjectionMap

from cil.utilities import dataexample
from testclass import CCPiTestClass
import scipy

from cil.utilities.errors import InPlaceError



initialise_tests()

def dt(steps):
    return steps[-1] - steps[-2]


class TestOperator(CCPiTestClass):


    def test_MatrixOperator(self):
        m = 30
        n = 20

        vg = VectorGeometry(n)

        Amat = numpy.random.randn(m, n)
        A = MatrixOperator(Amat)

        b = vg.allocate('random')

        out1 = A.range_geometry().allocate()
        out2 = A.domain_geometry().allocate()

        res1 = A.direct(b)
        res2 = numpy.dot(A.A, b.as_array())
        self.assertNumpyArrayAlmostEqual(res1.as_array(), res2)

        A.direct(b, out = out1)
        self.assertNumpyArrayAlmostEqual(res1.as_array(), out1.as_array(), decimal=4)

        res3 = A.adjoint(res1)
        res4 = numpy.dot(A.A.transpose(),res1.as_array())
        self.assertNumpyArrayAlmostEqual(res3.as_array(), res4, decimal=4)

        A.adjoint(res1, out = out2)
        self.assertNumpyArrayAlmostEqual(res3.as_array(), out2.as_array(), decimal=4)

        ## dot test for real and complex MatrixOperator
        self.assertTrue(A.dot_test(A))

        Amat = numpy.random.randn(m, n) + 1j*numpy.random.randn(m, n)
        A = MatrixOperator(Amat)
        self.assertTrue(A.dot_test(A))




    def test_ZeroOperator(self):
        ig = ImageGeometry(10,20,30)
        img = ig.allocate(3)
        out=ig.allocate(0)
        op1=ZeroOperator(ig)
        self.assertNumpyArrayEqual(op1.direct(img).array,out.array)
        self.assertNumpyArrayEqual(op1.adjoint(img).array,out.array)
        ig2 = ImageGeometry(10,15,30)
        out2=ig2.allocate(0)
        img2=ig2.allocate(5)
        op2=ZeroOperator(ig, ig2)
        self.assertNumpyArrayEqual(op2.direct(img).array,out2.array)
        self.assertNumpyArrayEqual(op2.adjoint(img2).array,out.array)
        self.assertEqual(op2.calculate_norm(),0)

    def test_ScaledOperator(self):
        ig = ImageGeometry(10,20,30)
        img = ig.allocate()
        scalar = 0.5
        sid = scalar * IdentityOperator(ig)
        numpy.testing.assert_array_equal(scalar * img.as_array(), sid.direct(img).as_array())


    def test_DiagonalOperator(self):
        M = 3
        ig = ImageGeometry(M, M)
        x = ig.allocate('random',seed=100)
        diag = ig.allocate('random',seed=101)

        # Set up example DiagonalOperator
        D = DiagonalOperator(diag)

        # Apply direct and check whether result equals diag*x as expected.
        z = D.direct(x)
        numpy.testing.assert_array_equal(z.as_array(), (diag*x).as_array())

        # Apply adjoint and check whether results equals diag*(diag*x) as expected.
        y = D.adjoint(z)
        numpy.testing.assert_array_equal(y.as_array(), (diag*(diag*x)).as_array())

        # test norm of diagonal
        norm1 = D.norm()
        numpy.testing.assert_almost_equal(norm1, numpy.max(diag.array))


    def test_DiagonalOperator_complex(self):
        M = 3
        ig = ImageGeometry(M, M)
        x = ig.allocate('random',seed=100 ,dtype=numpy.complex64)
        diag = ig.allocate('random',seed=101 ,dtype=numpy.complex64)

        # Set up example DiagonalOperator
        D = DiagonalOperator(diag)

        # Apply direct and check whether result equals diag*x as expected.
        z = D.direct(x)
        numpy.testing.assert_array_equal(z.as_array(), (diag*x).as_array())

        # Apply adjoint and check whether results equals diag.conjugate()*(diag*x) as expected.
        y = D.adjoint(z)
        numpy.testing.assert_array_equal(y.as_array(), (diag.conjugate()*(diag*x)).as_array())

        # test norm of diagonal with complex value
        norm1 = D.norm()
        numpy.testing.assert_almost_equal(norm1, numpy.max(numpy.abs(diag.array)))

    def test_MaskOperator(self):
        M = 3
        ig = ImageGeometry(M, M)
        x = ig.allocate('random',seed=100)

        mask = ig.allocate(True,dtype=bool)
        amask = mask.as_array()
        amask[2,1:3] = False
        amask[0,0] = False

        MO = MaskOperator(mask)

        # Apply direct and check whether result equals diag*x as expected.
        z = MO.direct(x)
        numpy.testing.assert_array_equal(z.as_array(), (mask*x).as_array())

        # Apply adjoint and check whether results equals diag*(diag*x) as expected.
        y = MO.adjoint(z)
        numpy.testing.assert_array_equal(y.as_array(), (mask*(mask*x)).as_array())


    def test_ChannelwiseOperator(self):
        M = 3
        channels = 4
        ig = ImageGeometry(M, M, channels=channels)
        igs = ImageGeometry(M, M)
        x = ig.allocate('random',seed=100)
        diag = igs.allocate('random',seed=101)

        D = DiagonalOperator(diag)
        C = ChannelwiseOperator(D,channels)

        y = C.direct(x)

        y2 = ig.allocate()
        C.direct(x,y2)

        for c in range(channels):
            numpy.testing.assert_array_equal(y.get_slice(channel=2).as_array(), \
                                             (diag*x.get_slice(channel=2)).as_array())
            numpy.testing.assert_array_equal(y2.get_slice(channel=2).as_array(), \
                                             (diag*x.get_slice(channel=2)).as_array())


        z = C.adjoint(y)

        z2 = ig.allocate()
        C.adjoint(y,z2)

        for c in range(channels):
            numpy.testing.assert_array_equal(z.get_slice(channel=2).as_array(), \
                                             (diag*(diag*x.get_slice(channel=2))).as_array())
            numpy.testing.assert_array_equal(z2.get_slice(channel=2).as_array(), \
                                             (diag*(diag*x.get_slice(channel=2))).as_array())


    def test_BlurringOperator(self):
        ig = ImageGeometry(100,100)

        # Parameters for point spread function PSF (size and std)
        ks          = 11;
        ksigma      = 5.0

        # Create 1D PSF and 2D as outer product, then normalise.
        w           = numpy.exp(-numpy.arange(-(ks-1)/2,(ks-1)/2+1)**2/(2*ksigma**2))
        w.shape     = (ks,1)
        PSF         = w*numpy.transpose(w)
        PSF         = PSF/(PSF**2).sum()
        PSF         = PSF/PSF.sum()

        # Create blurring operator
        BOP = BlurringOperator(PSF,ig)

        # Run dot test to check validity of adjoint.
        self.assertTrue(BOP.dot_test(BOP))


    def test_IdentityOperator(self):
        ig = ImageGeometry(10,20,30)
        img = ig.allocate()
        # img.fill(numpy.ones((30,20,10)))
        self.assertNumpyArrayEqual(img.shape , (30,20,10))
        #self.assertEqual(img.sum(), 2*float(10*20*30))
        self.assertEqual(img.sum(), 0.)
        Id = IdentityOperator(ig)
        y = Id.direct(img)
        numpy.testing.assert_array_equal(y.as_array(), img.as_array())

                
        #Check is_linear
        self.assertTrue(Id.is_linear())
        
        #Check is_orthogonal
        self.assertTrue(Id.is_orthogonal())

    def test_FiniteDifference(self):
        N, M = 2, 3
        numpy.random.seed(1)
        ig = ImageGeometry(N, M)
        Id = IdentityOperator(ig)

        FD = FiniteDifferenceOperator(ig, direction = 0, bnd_cond = 'Neumann')
        u = FD.domain_geometry().allocate('random')

        res = FD.domain_geometry().allocate(ImageGeometry.RANDOM)
        FD.adjoint(u, out=res)
        w = FD.adjoint(u)

        self.assertNumpyArrayEqual(res.as_array(), w.as_array())

        res = Id.domain_geometry().allocate(ImageGeometry.RANDOM)
        Id.adjoint(u, out=res)
        w = Id.adjoint(u)

        self.assertNumpyArrayEqual(res.as_array(), w.as_array())
        self.assertNumpyArrayEqual(u.as_array(), w.as_array())

        G = GradientOperator(ig)

        u = G.range_geometry().allocate(ImageGeometry.RANDOM)
        res = G.domain_geometry().allocate()
        G.adjoint(u, out=res)
        w = G.adjoint(u)

        self.assertNumpyArrayEqual(res.as_array(), w.as_array())

        u = G.domain_geometry().allocate(ImageGeometry.RANDOM)
        res = G.range_geometry().allocate()
        G.direct(u, out=res)
        w = G.direct(u)
        self.assertBlockDataContainerEqual(res, w)

        # 2D
        M, N = 2, 3
        ig = ImageGeometry(voxel_num_x=M, voxel_num_y=N, voxel_size_x=0.1, voxel_size_y=0.4)
        x = ig.allocate('random')

        labels = ["horizontal_y", "horizontal_x"]

        for i, dir in enumerate(labels):
            FD1 = FiniteDifferenceOperator(ig, direction=i)
            res1 = FD1.direct(x)
            res1b = FD1.adjoint(x)

            FD2 = FiniteDifferenceOperator(ig, direction=labels[i])
            res2 = FD2.direct(x)
            res2b = FD2.adjoint(x)

            numpy.testing.assert_almost_equal(res1.as_array(), res2.as_array())
            numpy.testing.assert_almost_equal(res1b.as_array(), res2b.as_array())

        # 2D  + chan
        M, N, K = 2,3,4
        ig1 = ImageGeometry(voxel_num_x=M, voxel_num_y=N, channels=K, voxel_size_x=0.1, voxel_size_y=0.4)
        x = ig1.allocate('random')

        labels = ["channel","horizontal_y", "horizontal_x"]

        for i, dir in enumerate(labels):
            FD1 = FiniteDifferenceOperator(ig1, direction=i)
            res1 = FD1.direct(x)
            res1b = FD1.adjoint(x)
            FD2 = FiniteDifferenceOperator(ig1, direction=labels[i])
            res2 = FD2.direct(x)
            res2b = FD2.adjoint(x)

            numpy.testing.assert_almost_equal(res1.as_array(), res2.as_array())
            numpy.testing.assert_almost_equal(res1b.as_array(), res2b.as_array())


    def test_PowerMethod(self):
        numpy.random.seed(2)
        # 2x2 real matrix, dominant eigenvalue = 2
        M1 = numpy.array([[1,0],[1,2]], dtype=float)
        M1op = MatrixOperator(M1)
        res1 = M1op.PowerMethod(M1op,100)
        numpy.testing.assert_almost_equal(res1,2., decimal=4)

        res_scipy = scipy.linalg.eig(M1)
        numpy.testing.assert_almost_equal(res1,numpy.abs(res_scipy[0]).max(), decimal=4)

        # Test with the norm
        res2 = M1op.norm()
        res1 = M1op.PowerMethod(M1op,100, method="composed_with_adjoint")
        numpy.testing.assert_almost_equal(res1,res2, decimal=4)


        # 2x3 real matrix, dominant eigenvalue = 4.711479432297657
        M1 = numpy.array([[1.,0.,3],[1,2.,3]])
        M1op = MatrixOperator(M1)
        res1 = M1op.PowerMethod(M1op,100)
        numpy.testing.assert_almost_equal(res1,4.711479432297657, decimal=4)

        res_scipy = scipy.linalg.eig(M1.T.conjugate()@M1)
        numpy.testing.assert_almost_equal(res1,numpy.sqrt(numpy.abs(res_scipy[0])).max(), decimal=4)

        # 2x3 complex matrix, (real eigenvalues), dominant eigenvalue = 5.417602365823937
        M1 = numpy.array([[2,1j,0],[2j,5j,0]])
        M1op = MatrixOperator(M1)
        res1 = M1op.PowerMethod(M1op,100)
        numpy.testing.assert_almost_equal(res1,5.531859582980837, decimal=4)
        res_scipy = scipy.linalg.eig(M1.T.conjugate()@M1)
        numpy.testing.assert_almost_equal(res1,numpy.sqrt(numpy.abs(res_scipy[0])).max(), decimal=4)


        # 3x3 complex matrix, (real+complex eigenvalue), dominant eigenvalue = 3.1624439599276974
        M1 = numpy.array([[2,0,0],[1,2j,1j],[3, 3-1j,3]])
        M1op = MatrixOperator(M1)
        res1 = M1op.PowerMethod(M1op,150)
        numpy.testing.assert_almost_equal(res1, 3.1624439599276974, decimal=3)
        res_scipy = scipy.linalg.eig(M1)
        print(numpy.abs(res_scipy[0]).max())
        numpy.testing.assert_almost_equal(res1,numpy.abs(res_scipy[0]).max(), decimal=4)

        # 2x2 non-diagonalisable nilpotent matrix
        M1=numpy.array([[0.,1.], [0.,0.]])
        M1op = MatrixOperator(M1)
        res1 = M1op.PowerMethod(M1op,5)
        numpy.testing.assert_almost_equal(res1,0, decimal=4)
        res_scipy = scipy.linalg.eig(M1)
        numpy.testing.assert_almost_equal(res1,numpy.abs(res_scipy[0]).max(), decimal=4)

        # 2x2 non-diagonalisable nilpotent matrix where method="composed_with_adjoint"
        M1=numpy.array([[0.,1.], [0.,0.]])
        M1op = MatrixOperator(M1)
        res1 = M1op.PowerMethod(M1op,5, method="composed_with_adjoint")
        numpy.testing.assert_almost_equal(res1,1, decimal=4)
        res_scipy = scipy.linalg.eig(M1.T@M1)
        numpy.testing.assert_almost_equal(res1,numpy.abs(res_scipy[0]).max(), decimal=4)


        # 2x2 matrix, max absolute eigenvalue is not unique and initial vector chosen for non-convergence

        M1=numpy.array([[2.,1.], [0.,-2.]])
        M1op = MatrixOperator(M1)
        _,_,_,_,convergence = M1op.PowerMethod(M1op,100, initial=DataContainer(numpy.array([1.,1.])), return_all=True)
        numpy.testing.assert_equal(convergence,False)

        # 2x2 matrix, max absolute eigenvalue is not unique and initial vector chosen for convergence

        M1=numpy.array([[2.,1.,0.],[0.,1.,1.], [0.,0.,1.]])
        M1op = MatrixOperator(M1)
        res1,_,_,_,convergence = M1op.PowerMethod(M1op,100, return_all=True)
        numpy.testing.assert_almost_equal(res1,2., decimal=4)
        numpy.testing.assert_equal(convergence,True)

        # Gradient Operator (float)
        ig = ImageGeometry(30,30)
        Grad = GradientOperator(ig)
        res1 = Grad.PowerMethod(Grad,500, tolerance=1e-6)
        numpy.testing.assert_almost_equal(res1, numpy.sqrt(8), decimal=2)

        # Gradient Operator (complex)
        ig = ImageGeometry(30,30, dtype=complex)
        Grad = GradientOperator(ig, backend='numpy')
        res1 = Grad.PowerMethod(Grad,500, tolerance=1e-6)
        numpy.testing.assert_almost_equal(res1, numpy.sqrt(8), decimal=2)

        # Identity Operator
        Id = IdentityOperator(ig)
        res1 = Id.PowerMethod(Id,100)
        numpy.testing.assert_almost_equal(res1,1.0, decimal=4)

        # Test errors produced if not a valid method
        try:
            res1 = Id.PowerMethod(Id,100, method='gobledigook')
        except ValueError:
            pass


    def test_Norm(self):
        numpy.random.seed(1)
        N, M = 200, 300

        ig = ImageGeometry(N, M)
        G = GradientOperator(ig)

        self.assertIsNone(G._norm)

        #calculates norm
        self.assertAlmostEqual(G.norm(), numpy.sqrt(8), 2)

        #sets_norm
        G.set_norm(4)
        self.assertEqual(G._norm, 4)

        #gets cached norm
        self.assertEqual(G.norm(), 4)

        #sets cache to None
        G.set_norm(None)
        #recalculates norm
        self.assertAlmostEqual(G.norm(), numpy.sqrt(8), 2)



        #Check that the provided element is a number or None
        with self.assertRaises(TypeError):
            G.set_norm['Banana']
        #Check that the provided norm is positive
        with self.assertRaises(ValueError):
            G.set_norm(-1)

         # 2x2 real matrix, dominant eigenvalue = 2. Check norm uses the right flag for power method
        M1 = numpy.array([[1,0],[1,2]], dtype=float)
        M1op = MatrixOperator(M1)
        res1 = M1op.norm()
        res2 = M1op.PowerMethod(M1op,100)
        res3 = M1op.PowerMethod(M1op,100, method="composed_with_adjoint")
        res4 = M1op.PowerMethod(M1op,100, method="direct_only")
        numpy.testing.assert_almost_equal(res1,res3, decimal=4)
        self.assertNotEqual(res1, res2)
        self.assertNotEqual(res1,res4)



    def test_ProjectionMap(self):
        # Check if direct is correct
        ig1 = ImageGeometry(3,4)
        ig2 = ImageGeometry(5,6)
        ig3 = ImageGeometry(5,6,4)

        # Create BlockGeometry
        bg = BlockGeometry(ig1,ig2, ig3)
        x = bg.allocate(10)

        # Extract containers
        x0, x1, x2 = x[0], x[1], x[2]

        for i in range(3):
            proj_map = ProjectionMap(bg, i)
            # res1 is in ImageData from the X_{i} "ImageGeometry"
            res1 = proj_map.direct(x)

            # res2 is in ImageData from the X_{i} "ImageGeometry" using out
            res2 = bg.geometries[i].allocate(0)
            proj_map.direct(x, out=res2)

            # Check with and without out
            numpy.testing.assert_array_almost_equal(res1.as_array(), res2.as_array())

            # Depending on which index is used, check if x0, x1, x2 are the same with res2
            if i==0:
                numpy.testing.assert_array_almost_equal(x0.as_array(), res2.as_array())
            elif i==1:
                numpy.testing.assert_array_almost_equal(x1.as_array(), res2.as_array())
            elif i==2:
                numpy.testing.assert_array_almost_equal(x2.as_array(), res2.as_array())
            else:
                pass

        # Check if adjoint is correct

        bg = BlockGeometry(ig1, ig2, ig3, ig1, ig2, ig3)
        x = ig1.allocate(20)

        index=3
        proj_map = ProjectionMap(bg, index)

        res1 = bg.allocate(0)
        proj_map.adjoint(x, out=res1)
        
        res2=bg.allocate('random')
        proj_map.adjoint(x, out=res2)

        # check if all indices return arrays filled with 0, except the input index

        for i in range(len(bg.geometries)):
            if i!=index:
                numpy.testing.assert_array_almost_equal(res1[i].as_array(), bg.geometries[i].allocate().as_array())
                numpy.testing.assert_array_almost_equal(res2[i].as_array(), bg.geometries[i].allocate().as_array())

        # Check error messages
        # Check if index is correct wrt length of Cartesian Product
        with self.assertRaises(ValueError):
            ig = ImageGeometry(3,4)
            bg = BlockGeometry(ig,ig)
            index = 3
            proj_map = ProjectionMap(bg, index)

        # Check error if an ImageGeometry is passed
        with self.assertRaises(ValueError):
            proj_map = ProjectionMap(ig, index)


class TestGradients(CCPiTestClass):
    def setUp(self):
        N, M = 20, 30
        K = 20
        C = 20
        self.decimal = 1
        self.iterations = 50
        ###########################################################################
        # 2D geometry no channels
        self.ig = ImageGeometry(N, M)
        self.ig2 = ImageGeometry(N, M, channels = C)
        self.ig3 = ImageGeometry(N, M, K)


    def test_SymmetrisedGradientOperator1a(self):
        ###########################################################################
        ## Symmetrized Gradient Tests
        ###########################################################################
        # 2D geometry no channels
        # ig = ImageGeometry(N, M)
        Grad = GradientOperator(self.ig)

        E1 = SymmetrisedGradientOperator(Grad.range_geometry())
        norm = LinearOperator.PowerMethod(E1, max_iteration=self.iterations)
        numpy.testing.assert_almost_equal(norm, numpy.sqrt(8), decimal = self.decimal)


    def test_SymmetrisedGradientOperator1b(self):
        ###########################################################################
        ## Symmetrized GradientOperator Tests
        ###########################################################################
        # 2D geometry no channels
        # ig = ImageGeometry(N, M)
        Grad = GradientOperator(self.ig)

        E1 = SymmetrisedGradientOperator(Grad.range_geometry())
        numpy.random.seed(1)
        u1 = E1.domain_geometry().allocate('random')
        w1 = E1.range_geometry().allocate('random', symmetry = True)

        lhs = E1.direct(u1).dot(w1)
        rhs = u1.dot(E1.adjoint(w1))
        # self.assertAlmostEqual(lhs, rhs)
        numpy.testing.assert_allclose(lhs, rhs, rtol=1e-3)


    def test_SymmetrisedGradientOperator2(self):
        ###########################################################################
        # 2D geometry with channels
        # ig2 = ImageGeometry(N, M, channels = C)
        Grad2 = GradientOperator(self.ig2, correlation = 'Space', backend='numpy')

        E2 = SymmetrisedGradientOperator(Grad2.range_geometry())
        numpy.random.seed(1)
        u2 = E2.domain_geometry().allocate('random')
        w2 = E2.range_geometry().allocate('random', symmetry = True)
    #
        lhs2 = E2.direct(u2).dot(w2)
        rhs2 = u2.dot(E2.adjoint(w2))

        numpy.testing.assert_allclose(lhs2, rhs2, rtol=1e-3)


    def test_SymmetrisedGradientOperator2a(self):
        ###########################################################################
        # 2D geometry with channels
        # ig2 = ImageGeometry(N, M, channels = C)
        Grad2 = GradientOperator(self.ig2, correlation = 'Space', backend='numpy')

        E2 = SymmetrisedGradientOperator(Grad2.range_geometry())
        norm = LinearOperator.PowerMethod(E2, max_iteration=self.iterations)
        numpy.testing.assert_almost_equal(norm,
           numpy.sqrt(8), decimal = self.decimal)


    def test_SymmetrisedGradientOperator3a(self):
        ###########################################################################
        # 3D geometry no channels
        #ig3 = ImageGeometry(N, M, K)
        Grad3 = GradientOperator(self.ig3, correlation = 'Space')

        E3 = SymmetrisedGradientOperator(Grad3.range_geometry())

        norm = LinearOperator.PowerMethod(E3,max_iteration=100, tolerance = 0)
        numpy.testing.assert_almost_equal(norm, numpy.sqrt(12), decimal = self.decimal)


    def test_SymmetrisedGradientOperator3b(self):
        ###########################################################################
        # 3D geometry no channels
        #ig3 = ImageGeometry(N, M, K)
        Grad3 = GradientOperator(self.ig3, correlation = 'Space')

        E3 = SymmetrisedGradientOperator(Grad3.range_geometry())
        numpy.random.seed(1)
        u3 = E3.domain_geometry().allocate('random')
        w3 = E3.range_geometry().allocate('random', symmetry = True)
    #
        lhs3 = E3.direct(u3).dot(w3)
        rhs3 = u3.dot(E3.adjoint(w3))

        numpy.testing.assert_almost_equal(lhs3, rhs3, decimal=3)
        self.assertTrue( LinearOperator.dot_test(E3, range_init = w3, domain_init=u3, decimal=3) )


    def test_dot_test(self):
        Grad3 = GradientOperator(self.ig3, correlation = 'Space', backend='numpy')

        # self.assertAlmostEqual(lhs3, rhs3)
        self.assertTrue( LinearOperator.dot_test(Grad3 , verbose=True, decimal=4))
        self.assertTrue( LinearOperator.dot_test(Grad3 , verbose=True, decimal=4))


    def test_dot_test2(self):
        Grad3 = GradientOperator(self.ig3, correlation = 'SpaceChannel', backend='c')

        # self.assertAlmostEqual(lhs3, rhs3)
        # self.assertTrue( LinearOperator.dot_test(Grad3 , verbose=True))
        self.assertTrue( LinearOperator.dot_test(Grad3 , decimal=4, verbose=True))



 

class TestOperatorCompositionSum(unittest.TestCase):
    def setUp(self):
        self.data = dataexample.BOAT.get(size=(128,128))
        self.ig = self.data.geometry


    def test_SumOperator(self):
        # numpy.random.seed(1)
        ig = self.ig
        data = self.data

        Id1 = 2 * IdentityOperator(ig)
        Id2 = IdentityOperator(ig)

        c = SumOperator(Id1,Id2)
        out = c.direct(data)

        numpy.testing.assert_array_almost_equal(out.as_array(),3 * data.as_array())


    def test_CompositionOperator_direct1(self):
        ig = self.ig
        data = self.data
        G = GradientOperator(domain_geometry=ig)

        Id1 = 2 * IdentityOperator(ig)
        Id2 = IdentityOperator(ig)

        d = CompositionOperator(G, Id2)

        out1 = G.direct(data)
        out2 = d.direct(data)

        numpy.testing.assert_array_almost_equal(out2.get_item(0).as_array(),  out1.get_item(0).as_array())
        numpy.testing.assert_array_almost_equal(out2.get_item(1).as_array(),  out1.get_item(1).as_array())


    def test_CompositionOperator_direct2(self):
        ig = self.ig
        data = self.data
        G = GradientOperator(domain_geometry=ig)


        Id1 = 2 * IdentityOperator(ig)
        Id2 = IdentityOperator(ig)

        d = CompositionOperator(G, Id2)

        out1 = G.direct(data)

        d_out = d.direct(data)

        d1 = Id2.direct(data)
        d2 = G.direct(d1)

        numpy.testing.assert_array_almost_equal(d2.get_item(0).as_array(),
                                                d_out.get_item(0).as_array())
        numpy.testing.assert_array_almost_equal(d2.get_item(1).as_array(),
                                                d_out.get_item(1).as_array())


    def test_CompositionOperator_direct3(self):
        ig = self.ig
        data = self.data
        G = GradientOperator(domain_geometry=ig)


        Id1 = 2 * IdentityOperator(ig)
        Id2 = IdentityOperator(ig)

        d = CompositionOperator(G, Id2)

        out1 = G.direct(data)

        d_out = d.direct(data)

        d1 = Id2.direct(data)
        d2 = G.direct(d1)

        numpy.testing.assert_array_almost_equal(d2.get_item(0).as_array(),
                                                d_out.get_item(0).as_array())
        numpy.testing.assert_array_almost_equal(d2.get_item(1).as_array(),
                                                d_out.get_item(1).as_array())

        G2Id = G.compose(2*Id2)
        d2g = G2Id.direct(data)

        numpy.testing.assert_array_almost_equal(d2g.get_item(0).as_array(),
                                                2 * d_out.get_item(0).as_array())
        numpy.testing.assert_array_almost_equal(d2g.get_item(1).as_array(),
                                                2 * d_out.get_item(1).as_array())


    def test_CompositionOperator_direct4(self):
        ig = self.ig
        data = self.data
        G = GradientOperator(domain_geometry=ig)

        Id1 = 2 * IdentityOperator(ig)
        Id2 = IdentityOperator(ig)

        d = CompositionOperator(G, Id1, Id2)

        out1 = G.direct(data)

        d_out = d.direct(data)

        numpy.testing.assert_array_almost_equal(d_out.get_item(0).as_array(),
                                                2 * out1.get_item(0).as_array())
        numpy.testing.assert_array_almost_equal(d_out.get_item(1).as_array(),
                                                2 * out1.get_item(1).as_array())


    def test_CompositionOperator_adjoint1(self):
        ig = self.ig
        data = self.data
        G = GradientOperator(domain_geometry=ig)

        Id1 = 2 * IdentityOperator(ig)
        Id2 = IdentityOperator(ig)

        d = CompositionOperator(G, Id2)
        da = d.direct(data)

        out1 = G.adjoint(da)
        out2 = d.adjoint(da)

        numpy.testing.assert_array_almost_equal(out2.as_array(),  out1.as_array())


    def test_CompositionOperator_adjoint2(self):
        ig = self.ig
        data = self.data
        G = GradientOperator(domain_geometry=ig)

        Id1 = 2 * IdentityOperator(ig)
        Id2 = IdentityOperator(ig)

        d = CompositionOperator(G, Id1)
        da = d.direct(data)

        out1 = G.adjoint(da)
        out2 = d.adjoint(da)

        numpy.testing.assert_array_almost_equal(out2.as_array(),  2 * out1.as_array())


    def test_CompositionOperator_adjoint3(self):
        ig = self.ig
        data = self.data
        G = GradientOperator(domain_geometry=ig)

        Id1 = 2 * IdentityOperator(ig)
        Id2 = IdentityOperator(ig)

        d = G.compose(Id1)
        da = d.direct(data)

        out1 = G.adjoint(da)
        out2 = d.adjoint(da)

        numpy.testing.assert_array_almost_equal(out2.as_array(),  2 * out1.as_array())


    def test_CompositionOperator_adjoint4(self):
        ig = self.ig
        data = self.data
        G = GradientOperator(domain_geometry=ig)


        Id1 = 2 * IdentityOperator(ig)

        d = G.compose(-Id1)
        da = d.direct(data)

        out1 = G.adjoint(da)
        out2 = d.adjoint(da)

        numpy.testing.assert_array_almost_equal(out2.as_array(),  - 2 * out1.as_array())


    def test_CompositionOperator_adjoint5(self):
        ig = self.ig
        data = self.data
        G = GradientOperator(domain_geometry=ig)

        Id1 = 3 * IdentityOperator(ig)
        Id = Id1 - IdentityOperator(ig)
        d = G.compose(Id)
        da = d.direct(data)

        out1 = G.adjoint(da)
        out2 = d.adjoint(da)

        numpy.testing.assert_array_almost_equal(out2.as_array(),  2 * out1.as_array())


    def test_CompositionOperator_adjoint6(self):
        ig = self.ig
        data = self.data
        G = GradientOperator(domain_geometry=ig)


        Id1 = 3 * IdentityOperator(ig)
        Id = ZeroOperator(ig)
        d = G.compose(Id)
        da = d.direct(data)

        out1 = G.adjoint(da)
        out2 = d.adjoint(da)

        numpy.testing.assert_array_almost_equal(out2.as_array(),  0 * out1.as_array())


    def stest_CompositionOperator_direct4(self):
        ig = self.ig
        data = self.data
        G = GradientOperator(domain_geometry=ig)

        sym = SymmetrisedGradientOperator(domain_geometry=ig)
        Id2 = IdentityOperator(ig)

        d = CompositionOperator(sym, Id2)

        out1 = G.direct(data)
        out2 = d.direct(data)

        numpy.testing.assert_array_almost_equal(out2.get_item(0).as_array(),  out1.get_item(0).as_array())
        numpy.testing.assert_array_almost_equal(out2.get_item(1).as_array(),  out1.get_item(1).as_array())


    def test_CompositionOperator_adjoint7(self):
        ig = self.ig
        data = self.data
        G = GradientOperator(domain_geometry=ig)

        Id1 = 2 * IdentityOperator(ig)
        Id2 = IdentityOperator(ig)

        d = CompositionOperator(G, Id1, Id2)

        out1 = G.direct(data)
        out2 = G.adjoint(out1)

        d_out = d.adjoint(out1)

        numpy.testing.assert_array_almost_equal(d_out.as_array(),
                                                2 * out2.as_array())
        numpy.testing.assert_array_almost_equal(d_out.as_array(),
                                                2 * out2.as_array())


    def test_CompositionOperator_norm1(self):

        M1 = numpy.array([[1,0],[1,2]], dtype=float)
        M1op = MatrixOperator(M1)

        M2=numpy.array([[3,1],[0,1]], dtype=float)
        M2op = MatrixOperator(M2)


        d = CompositionOperator(M1op, M2op)

        self.assertAlmostEqual(M2op.norm(), 3.1795867966025404,places=4)
        self.assertAlmostEqual(M1op.norm(), 2.288245580264566, places=4)
        self.assertAlmostEqual(d.norm(), 5.162277659459584, places=4)

    def test_CompositionOperator_norm2(self):

        M1 = numpy.array([[4,3],[1,1]], dtype=float)
        M1op = MatrixOperator(M1)

        M2=numpy.array([[1,-3],[-1,4]], dtype=float)
        M2op = MatrixOperator(M2)


        d = CompositionOperator(M1op, M2op)

        self.assertAlmostEqual(d.norm(), 1, places=4)


    def test_CompositionOperator_norm3(self):
        ig = self.ig
        M1 = numpy.array([[4,3],[1,1]], dtype=float)
        M1op = MatrixOperator(M1)

        M2=numpy.array([[1,-3],[-1,4]], dtype=float)
        M2op = MatrixOperator(M2)

        M3=numpy.array([[1,-5],[-1,6]], dtype=float)
        M3op = MatrixOperator(M2)

        d = CompositionOperator(M1op, M2op, M3op)

        out1 = M3op.norm()
        out2 = d.norm()

        numpy.testing.assert_almost_equal(out2, out1)


    def test_CompositionOperator_norm4(self):

        M1 = numpy.array([[4,0],[0,1]], dtype=float)
        M1op = MatrixOperator(M1)

        d = CompositionOperator(M1op)

        out1 = 4
        out2 = d.norm()

        numpy.testing.assert_almost_equal(out2, out1)



from cil.optimisation.algorithms import SIRT, GD, ISTA, FISTA
from cil.optimisation.functions import LeastSquares, IndicatorBox
from cil.framework import ImageGeometry, VectorGeometry
from cil.optimisation.operators import IdentityOperator, MatrixOperator

from cil.optimisation.utilities import Sensitivity, AdaptiveSensitivity, Preconditioner, ConstantStepSize, ArmijoStepSizeRule
import numpy as np

from testclass import CCPiTestClass
from unittest.mock import MagicMock


class TestStepSizes(CCPiTestClass):

    def test_step_sizes_called(self):

        ig = ImageGeometry(2, 1, 4)
        data = ig.allocate(1)
        A = IdentityOperator(ig)
        step_size_test = ConstantStepSize(3)
        step_size_test.get_step_size = MagicMock(return_value=.1)
        f = LeastSquares(A=A, b=data, c=0.5)
        alg = GD(initial=ig.allocate('random', seed=10), objective_function=f, step_size=step_size_test,
                 max_iteration=100, update_objective_interval=1)

        alg.run(5)

        self.assertEqual(len(step_size_test.get_step_size.mock_calls), 5)

        step_size_test = ConstantStepSize(3)
        step_size_test.get_step_size = MagicMock(return_value=.1)
        alg = ISTA(initial=ig.allocate('random', seed=10), f=f, g=IndicatorBox(lower=0), step_size=step_size_test,
                   max_iteration=100, update_objective_interval=1)
        alg.run(5)
        self.assertEqual(len(step_size_test.get_step_size.mock_calls), 5)

        step_size_test = ConstantStepSize(3)
        step_size_test.get_step_size = MagicMock(return_value=.1)
        alg = FISTA(initial=ig.allocate('random', seed=10), f=f, g=IndicatorBox(lower=0), step_size=step_size_test,
                    max_iteration=100, update_objective_interval=1)
        alg.run(5)
        self.assertEqual(len(step_size_test.get_step_size.mock_calls), 5)

    def test_constant(self):
        test_stepsize = ConstantStepSize(0.3)
        self.assertEqual(test_stepsize.step_size, 0.3)

    def test_armijo_init(self):
        test_stepsize = ArmijoStepSizeRule(alpha=1e3, beta=0.4, max_iterations=40)
        self.assertEqual(test_stepsize.alpha_orig, 1e3)
        self.assertEqual(test_stepsize.beta, 0.4)
        self.assertEqual(test_stepsize.max_iterations, 40)

        test_stepsize = ArmijoStepSizeRule()
        self.assertEqual(test_stepsize.alpha_orig, 1e6)
        self.assertEqual(test_stepsize.beta, 0.5)
        self.assertEqual(test_stepsize.max_iterations, np.ceil(
            2 * np.log10(1e6) / np.log10(2)))

    def test_armijo_calculation(self):
        test_stepsize = ArmijoStepSizeRule(alpha=8, beta=0.5, max_iterations=100)
        ig = VectorGeometry(2)
        data = ig.allocate('random')
        data.fill(np.array([3.5, 3.5]))
        A = MatrixOperator(np.diag([1., 1.]))
        f = LeastSquares(A, data)
        alg = GD(initial=ig.allocate(0), objective_function=f,
                 max_iteration=100, update_objective_interval=1, step_size=test_stepsize)
        alg.gradient_update = ig.allocate(-1)
        step_size = test_stepsize.get_step_size(alg)
        self.assertAlmostEqual(step_size, 4)

        alg.gradient_update = ig.allocate(-.5)
        step_size = test_stepsize.get_step_size(alg)
        self.assertAlmostEqual(step_size, 8)

        alg.gradient_update = ig.allocate(-2)
        step_size = test_stepsize.get_step_size(alg)
        self.assertAlmostEqual(step_size, 2)
